# MC_DSP.ps1
#From a Powershell window use Get-Help <PathToScript>\MC-DSP.ps1 to display Powershell help for this script.
#Use <PathToScript>\MC-DSP.ps1 -h or -help to display text help file.


<#
.synopsis
Exports or imports JRiver Media Center DSP preset registry keys to and from an xml file.
.description
Export function: reads JRiver Media Center DSP preset registry keys and properties then writes them to an xml file as custom objects.
Import function: reads an xml file and either creates a new Media Center DSP preset or overwrites the keys of an existing DSP preset with new properties.
Can be used to import specific DSP properties such as Parametric Equalizer filters.
Can be used to transfer Media Center DSP presets or properties between different systems or versions of Media Center.
.link 
https://yabb.jriver.com/interact/index.php/topic,124190.0.html
.notes
Revision 1.0 | 12th February 2020
Author: terrym@tassie
.parameter help
Displays the text file MC-DSP-HELP.txt
.parameter h
Synonym of help
.parameter export
Exports Media Center DSP preset keys from registry to an xml file
.parameter e
Synonym of export
.parameter import
Import Media Center DSP preset key(s) to registry from an xml file
.parameter i
Synonym of import
.parameter peq
Selects Parametric Equalizer (and if used) Parametric Equalizer 2 registry keys. Only valid with import function.
.parameter pq2
Selects keys from Parametric Equalizer and writes them to Parametric Equalizer 2. Only valid with import function.
.parameter list_presets
Lists the DSP preset names in the registry.
.parameter lp
Synonym of list_presets
.parameter delete_preset
Deletes DSP preset
.parameter dp
Synonym of delete_preset
.parameter list_keys
Lists the names of sub-keys in a DSP preset.
.parameter lk
Synonym of list_keys
.parameter list_file
Lists DSP preset name, DSP keys, Media Center version and system name from an xml file created by the export function.
.parameter lf
Synonym of list_file
.parameter verbose
Displays all property names being read by the export function or being created/overwritten by the import function.
Displays full registry paths used in import function.
.parameter v
Synonym of verbose
.parameter file
Name of file to be created by the export function or read from by the import and list_file functions.
Note: xml file extension (.xml) does not need to be included in file name.
.parameter f
Synonym of file
.parameter mc_ver
The version of Media Center specified for the export, list_presets or list_keys functions if multiple versions of Media Center are installed.
Note: Can optionally be specified for import function if transferring DSP presets between different Media Center versions.
.parameter mv
Synonym of mc_ver
.parameter preset
Name of a DSP preset for the export, list_keys and optionally for the import functions.
.parameter p
Synonym of preset
.parameter preset_key
Name of DSP preset sub-key(s), only specifiied sub-key(s) will be imported. Only valid with import function.
Names with spaces must be enclosed by quotes e.g "Room Correction"
Accepts multiple values, DSP preset sub-key names must be separated by comma (,) e.g Effects,"Room Correction"
.parameter pk
Synonym of preset_key
.parameter xml_path
Sets the path to folder where xml files created are stored by default. Default value is $PSScriptRoot
Relative addressing can be used e.g -XP dsp_files Sets default to <path to script>\dsp_files
Note: If the folder specified does not exist then XML_PATH defaults back to <path to script> 
.parameter xp
Synonym of xml_path
.example
If MC-DSP.ps1 is run without any parameters, MC-DSP prompts for the function to execute (export, import etc), then prompts for the additional parameters required for execution.
It is recommended to run MC-DSP without parameters (interactive mode) and use the menus and prompts for ease of use.
The following parameters can be specified and used with interactive mode:
XML_PATH, MC_VER and VERBOSE
If MC-DSP is run with just a function parameter such as -export then the user is prompted for: Media Center version, DSP preset name, Name (and optionally path) of xml file to be created.
Note: If -import is specified then user is only prompted for necessary parameters, optional parameters such as -PEQ, PQ2, PRESET and PRESET_KEY must be specified on command line. HINT: specify PRESET and PRESET_KEY as ? to get the same selection menus as interactive mode.
.example
To list all saved DSP preset names for Media Center 25
<path to script>\MC-DSP.ps1 -mc_ver 25 -list_presets

if only -list_presets is specified then MC-DSP prompts for Media Center version
.example
To list all the keys in a Media Center 25 preset named {my preset} 
<path to script>\MC-DSP.ps1 -mc_ver 25 -preset "my preset" -list_keys

if only -list_keys is specified then MC-DSP prompts for MC version and DSP preset name.
.example
To export all the properties for a Media Center 25 DSP preset named {my preset} to a file named {my exported preset.xml} 
<path to script>\MC-DSP.ps1 -export -mc_ver 25 -preset "my preset" -file "my exported preset"

if only -export is specified then MC-DSP prompts for MC version, DSP preset name and offers a default file name.
Note: default file name format is: preset name_MC version_computername.xml e.g mypreset_MC25_mysystem.xml
.example
To list the contents of an xml file created by MC-DSP named {my exported preset.xml} saved in the same folder as the script
<path to script>\MC-DSP.ps1 -list_file -file "my exported preset"

if only -list_file is specified then MC-DSP prompts for name of file.
The listing includes the name of the system the file was created on, version of Media Center that created the DSP preset, the DSP preset name, the registry path to the DSP preset and the registry keys of the DSP preset.
.example
To import all keys and properties from {my exported preset.xml} to the same MC version and leave DSP preset name unchanged
<path to script>\MC-DSP.ps1 -import -file "my exported preset"
.example
To import only the Parametric Equalizer properties from {my exported preset.xml} to a Media Center 26 DSP preset named {my peq} 
<path to script>\MC-DSP.ps1 -import -peq -mc_ver 26 -preset "my peq" -file "my exported preset"

if only -import -peq is specified then MC-DSP imports to same MC version as exported from (if installed) and prompts for name of file.
When the preset {my peq} is loaded using Media Center 26 DSP studio only the Parametric Equalizer settings will be changed.
.example
To import Parametric Equalizer properties to Parametric Equalizer 2 key
<path to script>\MC-DSP.ps1 -import -pq2 -mc_ver 26 -preset "my peq" -file "my exported preset"

The properties of Parametric Equalizer in file "my exported preset" will be written to the Parametric Equalizer 2 key of preset "my peq"
.example
To import only the {Room Correction} properties from {my exported preset.xml} to a preset named {my imported RC} 
<path to script>\MC-DSP.ps1 -import -preset_key "Room Correction"  -preset "my imported RC" -file "my exported preset"

When the preset {my imported RC} is loaded using Media Center DSP studio only the {Room Correction} settings will be changed.
As in the previous example if the file was created using MC25 and the target version is MC26 then also specify -mc_ver 26
#>


param(
    [switch]$HELP,[switch]$H,[switch]$DELETE_PRESET,[switch]$DP,
    [switch]$EXPORT,[switch]$E,[switch]$IMPORT,[switch]$I,[switch]$PEQ,[switch]$PQ2,
    [switch]$LIST_PRESETS,[switch]$LP,[switch]$LIST_KEYS,[switch]$LK,[switch]$LIST_FILE,[switch]$LF,[switch]$VERBOSE,[switch]$V,
    [string]$FILE,[string]$F,[string]$MC_VER,[string]$MV,[string]$PRESET,[string]$P,[array]$PRESET_KEY,[array]$PK,
    [string]$XML_PATH=$PSScriptroot,[string]$XP
    )

# uncomment (remove #) and edit the line below to set a default Media Center version
#e.g if you only use JRiver MC25 then $MC_VER="25"
#$MC_VER="25"

#Uncomment and edit line below to change the default folder used to save or search for xml files
#e.g to use a sub-folder of the script location named dsp_files edit variable below to $XML_PATH="$PSScriptroot\dsp_files"
#e.g to use sub=folder of Documents edit variable to $XML_PATH="$home\Documents\dsp_files"
#$XML_PATH=$PSScriptroot

#parse synonyms
if($H){$HELP=$true};if($DP){$DELETE_PRESET=$true};if($E){$EXPORT=$true};if($I){$IMPORT=$true}if($LP){$LIST_PRESETS=$true}
if($LK){$LIST_KEYS=$true};if($LF){$LIST_FILE=$true};if($V){$VERBOSE=$true};if($F){$FILE=$F};if($MV){$MC_VER=$MV};if($P){$PRESET=$P}
if($PK){$PRESET_KEY=$PK};if($XP){$XML_PATH=$XP}


Function Set-MC_VER{
    #if MC revision is not specified and required, ask for input and return string: empty, "Q" or valid MC_VER
    #Parameters: -rv check for installed versions, -select_msg specifies message shown before MC_VER input
    param([switch]$rv,[string]$select_msg="Please select a Media Center version")
    $installed_mc=Get-ChildItem "HKCU:\Software\J. River" -Name
    $vers=($installed_mc -join ", ") -replace "Media Center "
    #if $vers empty then return empty string else return comma separated string of versions
    if($rv -or (-not$vers)){
        return $vers
    }
    #if only single installed MC version then return $vers, ignore $MC_VER
    if($vers -match "^\d\d$"){
        return $vers
    }
    #if multiple MC versions installed, MC_VER is a valid number and matches a version return $MC_VER 
    elseif($MC_VER -match "^\d\d$"){
        if($vers -match $MC_VER){
            return $MC_VER
        }
        else{
            #need user input to resolve if MC_VER not match an installed version
            Write-Host "INFO:`tParameter <MC_VER> specifies Media Center $MC_VER"
            Write-Host "ERROR:`tMedia Center $MC_VER is not installed on this system!" -ForegroundColor Red
        }
    }
    #if MC_VER not set or invalid, display available versions and get user input
    if($MC_VER){
        Write-Host "WARN:`tParameter <MC_VER> is set but contains an invalid value: $MC_VER"
    }
    Write-Host "INFO:`tInstalled Media Center versions: $vers"
    Write-Host "INFO:`t$select_msg"
    do{
        $mcv=Read-Host -Prompt "INPUT:`tInput [Q] to QUIT <or> 
                                    `r`tInput [VERSION NUMBER] of Media Center (e.g $vers)"
        if($mcv -eq "q"){return $mcv}
        if($vers -notmatch $mcv){
            Write-Host "WARN:`tPlease input a Media Center version that is installed!"
            Write-Host "INFO:`tInstalled Media Center versions: $vers"
        }
    }until($mcv -match "^\d\d$" -and ($vers -match $mcv))
    Write-Host ""
    return $mcv

}#End Function Set-MC_VER


Function Set-Preset_mcv{
    #Sets MC version for preset functions returns MC version or null string
    #Parameter: $mc_msg is message to be displayed in Set-MC_VER function
    param($mc_msg)
    $mcv=Set-MC_VER -select_msg $mc_msg
    if(-not$mcv -or $mcv -eq "q"){
        if($mcv -eq "q"){Write-Host "`nERROR:`tNO MC version selected! Quitting!`n" -ForegroundColor Red}
        if(-not$mcv){Write-Host "`nERROR:`tNO MC version installed! Quitting!`n" -ForegroundColor Red}
        $script:FLAG_ERROR=$true
        return
    }
    return $mcv
}


Function Set-DSP_names{
    #function sets and validates either DSP preset names or if -pk DSP key names
    #Parameters: -valid_names array of presets or preset keys, -clp command line parameter to validated, -pk specifies preset_keys
    #Accepts array input if -pk and returns array else input and return are strings
    #$msg is modifer for input prompt, inherits $mcv, $file_preset from parent function
    param([array]$valid_names,$clp,[switch]$pk,$msg)
    if($pk){
        $prompt="`nINPUT:`tPlease input [DSP preset key name(s)] $msg
        Multiple values must be separated by a comma (,)
        (Input [Q] to QUIT <or> [L] to list key names)`n
        Input [Q] <or> [L] <or> [DSP key name(s)]"
        $list_msg="`nINFO:`t$file_preset keys:`n"
        $err_msg="`nERROR:`tNO DSP preset key named:"
    }
    else{
        $prompt="`nINPUT:`tPlease input [DSP preset name] $msg
        (Input [Q] to QUIT <or> [L] to list preset names)`n
        Input [Q] <or> [L] <or> [DSP preset name]"
        $list_msg="`nINFO:`tMC$mcv presets:`n"
        $err_msg="`nERROR:`tNO DSP preset named:"
    }
    $pass=1
    do{
        #command line parameter(clp) not specified or clp was invalid and new input required
        if($pass -ne 1 -or -not$clp -or $clp -eq "?"){
            $clp=Read-Host -Prompt $prompt
            if($clp -eq "q"){return $null}
            if($clp -eq "l"){
                Write-Host $list_msg
                $valid_names.foreach{Write-Host "`t$_"}
                $pass++
                continue
            }
        }
        $valid_entry=@()
        $fix=@()
        #fix common input errors
        if($clp -match ","){$clp=$clp -split ","}
        $fix+=$clp.foreach{($_ -replace "`"|`'").Trim()}
        foreach($p in $fix){
            $p=[Regex]::Escape($p)
            $entry=$valid_names -match "^$p$"
            if($entry){
                $valid_entry+=$entry
            }
            else{
                $clp_err=[Regex]::Unescape($p)
                break
            }
        }
        if($clp_err){
            #if error warn then list names
            Write-Host $err_msg $clp_err -ForegroundColor Red
            Write-Host $list_msg
            $valid_names.foreach{Write-Host "`t$_"}
            Clear-Variable -name clp_err,valid_entry
            $pass++
            continue
        }
    }until ($valid_entry)
    if($pk){
        return $valid_entry
    }
    return [string]$valid_entry
    
}#end of Function Set-DSP_names


Function Set-Valid_preset{
    #checks or sets valid preset name, returns DSP preset name or null string
    #Parameters: $mcv is MC version, $vp_msg is message modifier for Set-DSP_names
    param($mcv,$vp_msg)
    #get presets in array for Set-DSP_names
    [array]$preset_names=Get-ChildItem "HKCU:\Software\J. River\Media Center $mcv\Zone Presets" -Name
    #if no presets warn and quit
    if($preset_names.Length -eq 0){
        Write-Host "`nERROR:`tNO DSP presets found for Media Center $mcv! Quitting function" -ForegroundColor Red
        $script:FLAG_ERROR=$true
        return
    }
    #if parameter PRESET check validity
    if($PRESET){
        $valid_preset=Set-DSP_names -valid_names $preset_names -clp $PRESET -msg $vp_msg
    }
    else{
        $valid_preset=Set-DSP_names -valid_names $preset_names -msg $vp_msg
    }   
    if(-not$valid_preset){
        Write-Host "`nERROR:`tNO DSP preset selected! Quitting function" -ForegroundColor Red
        $script:FLAG_ERROR=$true
        return
    }
    return $valid_preset
}#end of function Set-Valid_preset


Function Set-Export_params {
    #Function checks and optionally gathers user input to pass to Export-DSP function which exports registry keys to xml file

    #if $MC_VER set then check validity if not set then ask
    $mcv=Set-Preset_mcv -mc_msg "Please select a version to export a DSP preset from:"
    if(-not$mcv){return}
    $valid_preset=Set-Valid_preset -mcv $mcv -vp_msg "to export"
    if(-not$valid_preset){return}
    $dsp_path="HKCU:\Software\J. River\Media Center $mcv\Zone Presets\$valid_preset"
    $path=$XML_PATH
    #use save file gui dialog even if -file specified to use save file error checking e.g file exists, illegal name
    if($FILE){
        #filename specified by -file
        #force '.xml' extension to file name and if path specified update $path
        $title="MC-DSP EXPORT: Please confirm specified filename. Click Save to create file or Cancel to quit"
        if($FILE -notmatch "\.xml$"){
            $script:FILE="$FILE.xml"
        }
        if($FILE -match "\\"){
            $path=$FILE -replace "\\[^\\]+$"
            $script:FILE=$FILE -replace "^.+\\"
        }
    }
    else{
        #if file name not specified by -file, offer default file name and give user option to enter filename of choice
        #default file name format: <preset name_MC version_computername.xml> e.g mypreset_MC25_mysystem.xml
        #make sure file name does not contain illegal characters
        $script:FILE=(($valid_preset + "_MC$mcv" + "_$env:COMPUTERNAME.xml") -replace "\s","_") -replace "\\|/|:|,|\*|\?|`"|<|>|\||!|`'","-"
        $title="MC-DSP EXPORT: Please enter a filename or accept default. Click Save to create file or Cancel to quit"
    }
    #if the user specified path or XML_PATH does not exist default to $PSScriptroot for save file dialog
    if(-not(Test-Path $path)){
        $path=$PSScriptroot
    }    
    $FileSaver = New-Object System.Windows.Forms.SaveFileDialog -Property @{ 
        InitialDirectory = $path
        Title = $title
        Filter = 'XML files(*.xml)|*.xml'
    }
    $FileSaver.filename=$FILE
    $fsr=$FileSaver.ShowDialog()
    if($fsr -eq "OK"){
        $filepath=$FileSaver.filename
    }
    else{
        Write-Host "WARN:`tFilename not specified! Quitting Export function"
        return
    }

    #read registry and create xml file
    Get-Childitem $dsp_path|Export-DSP -xml_file $filepath
    if($FLAG_ERROR){
        return
    }
    Write-Host "`nINFO:`tCompleted export of MC$mcv DSP preset: $valid_preset"
}#End of Function Set-Export_params

Function Export-DSP{

    #Export registry keys to xml file.
    #Parameters: accepts pipeline input for $dsp_preset, -xml_file specifies output filepath
    [cmdletBinding()]
    param(
        [Parameter(Position=0,Mandatory=$True,
        HelpMessage="Enter a registry path using the PSDrive format.",
        ValueFromPipeline=$True,ValueFromPipelineByPropertyName=$True)]
        [ValidateScript({(Test-Path $_) -AND ((Get-Item $_).PSProvider.Name -match "Registry")})]
        [Alias("PSPath")]
        [string]$dsp_preset,
        [Parameter()]
        [string]$xml_file
    )
    Begin{
        Function Write-reg_object{
            #create custom object for registry entries
            $reg_object=New-Object -TypeName PSObject -Property @{
                "Path"=$item
                "Name"=$property_item
                "Value"=$value
                "Type"=$reg_type
                "Computername"=$env:computername
                }
            return $reg_object
        }#end function Write-Object
        
        Write-Host "`nINFO:`tReading properties from MC$mcv DSP preset: $valid_preset`n"
        #initialize an array to hold the results
        $data=@()
    }
    Process{
        #go through each pipelined path
        foreach ($item in $dsp_preset){
            #ensure $item is formatted as HKCU: for consistency
            $item=$item -replace "^.+_USER","HKCU:"
            if($VERBOSE){$pk_msg=$item}else{$pk_msg=$item -replace "^.+\\"}
            Write-Host "INFO:`tProcessing key - $pk_msg"
            $reg_item=Get-Item -Path $item
            #get property names
            $properties= $reg_item.Property
            
            if (-not ($properties)){
                    #no item properties were found so create a default entry
                    #create a custom object for entry and add it to array
                    Write-Host "`tNo properties found, creating default entry"
                    $property_item="(Default)"
                    $value=$null
                    $reg_type="String"
                    $data+=Write-reg_object
                }       
                else{
                    Write-Host "`t`tReading $(($properties | measure-object).count) properties"
                    #enumerate each property getting name,value and type
                    foreach ($property in $properties) {
                        if($VERBOSE){
                            Write-Host "`t`t$property"
                        }
                        $value=$reg_item.GetValue($property,$null,"DoNotExpandEnvironmentNames")
                        #get the registry value type
                        $reg_type=$reg_item.GetValueKind($property)
                        $property_item=$property
                        #create a custom object for entry and add it to array
                        $data+=Write-reg_object
                    }
                }
        }#close foreach 
    }#close Process
    End{
        #make sure we have something to write to file
        if ($data){
            #export to file
            Write-Host "`nINFO:`tExporting properties to file:`n`t$xml_file"
            try{
                $data | Export-CLIXML -Path $xml_file
            }
            catch{
                $PSerror=$error[0]
                Write-Host "ERROR:`tFailed to write XML file:`n`t$xml_file`n`tError message is:`n`t$PSerror" -ForegroundColor Red
                $script:FLAG_ERROR=$true
                return
            }
        }
        else{
            Write-Host "ERROR:`tNo properties read from registry! - Unable to create XML file!" -ForegroundColor Red
            $script:FLAG_ERROR=$true
            return
        }
    }#close End
} #End Function Export-DSP


Function Set-Import_mcv{
    #Function sets MC version for Import and returns hashtable $sim
    #inherits $rx and $preset_root from parent Import-DSP function
    [HashTable]$sim=@{}
    $file_mc_ver=$rx.file_mc_ver
    $f_ver=$file_mc_ver -replace ("\D")
    #get installed MC versions
    $vers=Set-MC_VER -rv
    if(-not$vers){
        Write-Host "`nERROR:`tNO Media Center version is installed! Quitting Import function`n" -ForegroundColor Red
        $script:FLAG_ERROR=$true
        $sim.nomc=$true
        return $sim
    }
    #single or multi installed MC case
    #case 1 file NOT match but $MC_VER valid and matches $vers: mod path and return
    if($MC_VER -ne $f_ver -and $MC_VER -match "^\d\d$" -and $vers -match $MC_VER){
        Write-Host "WARN:`tDSP preset was created by:      Media Center $f_ver
                  `rINFO:`tParameter <MC_VER> specifies:   Media Center $MC_VER
                  `rINFO:`tDSP preset will be imported to: Media Center $MC_VER"
        $sim.preset_root=$preset_root -replace "Center[\s\d]+","Center $MC_VER"
        $sim.mcv=$MC_VER
        return $sim
    }
    
    #single installed MC version cases
    if($vers -match "^\d\d$"){
        #case 2 single mc, file matches, no user dialogue needed, ignore MC_VER, no mods to path
        if($vers -eq $f_ver){
            $sim.nomod=$true
            $sim.mcv=$f_ver
            return $sim
        }
        #case 3 single mc, file NOT match, no matching $MC_VER: ASK to change version on import or quit
        if($vers -ne $f_ver -and $vers -ne $MC_VER){
            Write-Host "WARN:`tDSP preset was created by:          $file_mc_ver"
            if($MC_VER -match "^\d\d$"){
                Write-Host "INFO:`tParameter mc_ver is set to:         Media Center $MC_VER"
            }
            Write-Host "WARN:`tDSP Preset can only be imported to: Media Center $vers"
            do{
                Write-Host "QUERY:`tPress [Q] to QUIT <or> [ENTER] key to continue IMPORT:" -NoNewline
                $kp=[console]::Readkey($true)
                Write-Host "`n"
                if($kp.KeyChar -eq "q"){
                    Write-Host "INFO:`tUser quit Import function"
                    $sim.quit=$true
                    return $sim
                }
            }until($kp.KEY -eq "Enter")
            Write-Host "INFO:`tImporting DSP preset to Media Center $vers"
            $sim.preset_root=$preset_root -replace "Center[\s\d]+","Center $vers"
            $sim.mcv=$vers
            return $sim
        }
    }#end of single mc cases

    #multi installed MC version cases
    if($vers -match ","){
        #valid MC_VER specified (number and matches an installed version)
        if($MC_VER -match "^\d\d$"-and $vers -match $MC_VER -and $MC_VER -eq $f_ver){
            #case 4: MC_VER matches f_ver and vers: return preset root
            $sim.nomod=$true
            $sim.mcv=$f_ver
            return $sim
        }
        #invalid or no MC_VER -interactive cases
        #case 5 MC_VER NOT matches vers: warn and clear MC_VER
        if($MC_VER -match "^\d\d$" -and $vers -notmatch $MC_VER){
                Write-Host "INFO:`tParameter <MC_VER> specifies:       Media Center $MC_VER
                      `nWARN:`tMedia Center $MC_VER is NOT installed!`n"
            $Script:MC_VER=""
            $c5=$true
        }
        #case 6 MC_VER is not set/invalid: output file ver, installed vers and use Set-MC_VER for user input
        #return either preset_root. preset_root mod or quit to import
        if($MC_VER -notmatch "^\d\d$"){
            Write-Host     "INFO:`tDSP preset was created by:          $file_mc_ver"
            #if file version matches an installed version then offer this as default choice
            if($vers -match $f_ver){
                Write-Host "INFO:`tDSP Preset can be imported to:      Media Center $vers"
                do{
                    Write-Host "QUERY:`tPress [ENTER] to accept default of: " -NoNewline
                    Write-Host "$file_mc_ver" -NoNewline -ForegroundColor Yellow
                    Write-Host " <or>`n`tPress [S] key to SELECT Media Center version for import:" -NoNewline
                    $kp=[console]::Readkey($true)
                    Write-Host "`n"
                    if($kp.key -eq "Enter"){
                        Write-Host "INFO:`tImporting DSP preset to $file_mc_ver"
                        $sim.nomod=$true
                        $sim.mcv=$f_ver
                        return $sim
                    }
                    if($kp.KeyChar -eq "s"){
                        $mcv=Set-MC_VER -select_msg "Please select a Media Center version for DSP import"
                        if($mcv -eq "q"){
                            Write-Host "`nINFO:`tUser quit Import function"
                            $sim.quit=$true
                            return $sim
                        }
                        Write-Host "`nINFO:`tImporting DSP preset to Media Center $mcv"
                        $sim.preset_root=$preset_root -replace "Center[\s\d]+","Center $mcv"
                        $sim.mcv=$mcv
                        return $sim
                    }
                }until ($mcv)
            }
            else{
                #file version not match any installed version so go straight to MC version select
                if(-not$c5){Write-Host "`nWARN:`tMedia Center $f_ver is NOT installed!`n"}
                $mcv=Set-MC_VER -select_msg "Please select a Media Center version for DSP import"
                if($mcv -eq "q"){
                    Write-Host "`nINFO:`tUser quit Import function"
                    $sim.quit=$true
                    return $sim
                }
                Write-Host "`nINFO:`tImporting DSP preset to Media Center $mcv"
                $sim.preset_root=$preset_root -replace "Center[\s\d]+","Center $mcv"
                $sim.mcv=$mcv
                return $sim
            }#end of else
        }#end of if
    }#end of multi version cases

    $sim.quit=$true
    return $sim  
}#end of function Set-Import_mcv


Function Set-Import_scope {
    #Inherits $rx from parent Import-DSP function
    #Query user for scope of import: PEQ, Specific keys, All or copy PEQ to PEQ2
    #Sets script variables $PEQ, $PRESET_KEY
    do{            
        Write-Host "`n`tIMPORT SCOPE MENU - Press a [key] to select which DSP keys to import`n"       
        Write-Host "[P] PEQ  `t[S] SPECIFIC`t[2] PEQ=>PEQ2"
        Write-Host "[A] ALL  "  -ForegroundColor Yellow -NoNewline
        Write-Host "`t[Q] QUIT`t[?] Help`t(default is `"A`"):" -NoNewline
        $kp=[console]::Readkey($true)
        Write-Host "`r"
        if($kp.KeyChar -eq "q"){
            return "q"
        }
        if($kp.KeyChar -eq "p"){
             #validate keys available
            if($rx.dsp_key_array -match "Parametric"){
                $script:PEQ=$true
                $exit_menu=$true
            }
            else{
                Write-Host "`nERROR:`tPEQ selected but file does not contain any Parametric Equalizer keys!" -ForegroundColor Red
            }
        }
        if($kp.KeyChar -eq "2"){
            #validate keys available
           if($rx.dsp_key_array -match "Parametric Equalizer$"){
               $script:PQ2=$true
               $script:PEQ=$true
               $exit_menu=$true
           }
           else{
               Write-Host "`nERROR:`tPEQ=>PEQ2 selected but file does not contain Parametric Equalizer key!" -ForegroundColor Red
           }
       }
        if($kp.KeyChar -eq "s"){
            $script:PRESET_KEY=Set-DSP_names -valid_names $rx.dsp_key_array -pk
            if($PRESET_KEY){
                $exit_menu=$true
            }
            else{
                Write-Host "`nERROR:`tNo DSP preset key names were entered" -ForegroundColor Red
            }
        }
        if($kp.KeyChar -eq "a" -or $kp.Key -eq "Enter") {
            $exit_menu=$true
        }
        if($kp.KeyChar -eq "?"){
            Write-Host "`n[A] ALL `tPress [A] to import all DSP keys"
            Write-Host "[P] PEQ `tPress [P] to import only Parametric Equalizer (PEQ) keys"
            Write-Host "[2] PEQ=>PEQ2`tPress [2] to import Parametric Equalizer(PEQ) to Parametric Equalizer 2(PEQ2)"
            Write-Host "[S] SPECIFIC`tPress [S] to choose specific DSP preset key(s) to import"
            Write-Host "[Q] QUIT`tPress [Q] to terminate Import and return to Main Menu"
            Write-Host "[?] Help`tPress [?] to display help about menu choices"
            Write-Host "(default)`tPress [ENTER] to accept default choice of [A] ALL KEYS`n"
        }
    }until($exit_menu)
}# end of function Set-Import_scope

Function Read-XML_file{
    #Function reads data from xml file & displays MC version that created DSP, system name, DSP name, DSP keys &
    # registry path. Returns a hash table containing array of objects read from file, array of DSP key names, 
    # MC version string and DSP registry path string.
    [hashtable]$rx=@{}
    $path=$XML_PATH
    if($FILE){
        if($FILE -notmatch "\.xml$"){
            $script:FILE="$FILE.xml"
        }
        if($FILE -match "\\"){
            $filepath=$FILE
            $path=$FILE -replace "\\[^\\]+$"
            $script:FILE=$FILE -replace "^.+\\"
        }
        else{
            $filepath="$XML_PATH\$FILE"
            $path=$XML_PATH
        }
        if(-not(Test-Path $filepath)){
            $gui=$true
            $title="MC-DSP:  Please confirm specified file to Open - Click Cancel to Quit"
            $shelp=$true
        }
    }
    else{
        $title="MC-DSP:  Please select file to Open - Click Cancel to Quit"
        $gui=$true
    }
    #use gui if no file specified or file not found
    if($gui){
        #if path not valid default to PSScriptroot
        if(-not(Test-Path $path)){$path=$PSScriptroot}
        $FileBrowser = New-Object System.Windows.Forms.OpenFileDialog -Property @{ 
            InitialDirectory = $path
            Title = $title
            Filter = 'XML files(*.xml)|*.xml'
        }
        #due to MS bug need to use old style filebrowser to display file if specified -new style misaligns name in filename box
        #setting ShowHelp true uses old style
        $FileBrowser.ShowHelp=$shelp
        $FileBrowser.FileName=$FILE
        $fbr=$FileBrowser.ShowDialog()
        if($fbr -eq "OK"){
            $filepath=$FileBrowser.filename
        }
        else{
            Write-Host "WARN:`tNo file selected - Quitting"
            return
        }
    }#end of gui
    #read the file into array
    try{
        Write-Host "`nINFO:`tReading xml file:`n`t$filepath"
        $rx.data=Import-Clixml $filepath
    }
    catch{
        $PSerror=$error[0]
        Write-Host "ERROR:`tFailed to read XML file:`n`t$filepath`n`tError message is:`n`t$PSerror" -ForegroundColor Red
        Write-Host "HELP:`tPossible corrupt file, also check file was created by MC-DSP"
        $script:FLAG_ERROR=$true
        return
    }
    #Check we have data we can work with
    if($rx.data[0].Path -notmatch "\\Zone Presets\\"){
        Write-Host "ERROR:`tData read from file is not usable! Corrupt xml file!" -ForegroundColor Red
        Write-Host "HELP:`tCorrupt xml data, check file was created by MC-DSP"
        $script:FLAG_ERROR=$true
        return
    }
    #Display name of system the file was created on, the MC version, the DSP preset name, registry path and keys
    #extract 'root' DSP key path
    $rx.preset_root=$rx.data[0].path -replace "\\[^\\]+$"
    $rx.file_preset=$rx.preset_root -replace "^.+\\"
    $rx.file_mc_ver=$rx.preset_root -replace "^.+River\\|\\Zone.+$"
    $computername=$rx.data[0].computername
    Write-Host "`nINFO:`tFile was created on system named: $computername"
    Write-Host "INFO:`tDSP preset was created by: $($rx.file_mc_ver)"
    Write-Host "INFO:`tExport DSP registry path:`n`t$($rx.preset_root)"
    Write-Host "INFO:`tDSP preset name is: $($rx.file_preset)"
    Write-Host "INFO:`t$($rx.file_preset) keys:`n"
    
    #put all dsp_key names in array to be used for input validity checks
    $dsp_key_array=@()
    foreach($item in $rx.data){
        if($prev-ne $item.path){
            $dsp_key=$item.path -replace "^.+\\"
            Write-Host "`t$dsp_key"
            $dsp_key_array+=$dsp_key
        }
        $prev=$item.path
    }
    if($LIST_FILE){Write-Host "`nINFO:`tCompleted file listing";return}
    $rx.dsp_key_array=$dsp_key_array
    return $rx
}#end of Function Read-XML_file


Function Import-DSP {
    #Function validates import MC version, allows selection of DSP keys & DSP name.
    #Writes DSP settings to MC registry keys. 
    #Calls functions Read-XML_file, Set-Import_mc and if interactive Set-Import_scope.
    
    #Get xml file data
    $rx=Read-XML_file
    if(-not$rx){return}
    $file_preset=$rx.file_preset
    $preset_root=$rx.preset_root
    Write-Host ""
    #Set import MC version
    $sim=Set-Import_mcv -file_mc_ver $rx.file_mc_ver -preset_root $preset_root
    if($sim.nomc){return}
    if($sim.quit){return}
    if($sim.nomod){$mcv=$sim.mcv}
    if($sim.preset_root){
        $preset_root=$sim.preset_root
        $mcv=$sim.mcv
        $path_mod=$true
    }
    #if pq2 specified set peq
    if($PQ2){$script:PEQ=$true}
    #if -peq or -preset_key specified on command line check file contains relevant key path
    if($PEQ -and -not($rx.dsp_key_array -match "Parametric")){
        Write-Host "ERROR:`t-peq or -pq2 specified but selected file does not contain any Parametric Equalizer keys!" -ForegroundColor Red
        Write-Host "HELP:`tIf -peq or -pq2 used with -import specifier, selected file must contain Parametric Equalizer keys"
        $script:FLAG_ERROR=$true
        return
    }
    if($PRESET_KEY){
        #validate preset_key from command line
        $script:PRESET_KEY=Set-DSP_names -valid_names $rx.dsp_key_array -clp $PRESET_KEY -pk
        if(-not$PRESET_KEY){
            #offer abort or continue with all keys being imported
            Write-Host "WARN:`tNO DSP preset keys were selected!"
            Write-Host "WARN:`tAll DSP preset keys contained in file will be imported"
            do{
                Write-Host "QUERY:`tPress [Q] to QUIT <or> Press [ENTER] to continue import:" -NoNewline
                $kp=[console]::Readkey($true)
                Write-Host "`r"
                if($kp.KeyChar -eq "q"){
                    Write-Host "INFO:`tUser quit Import function"
                    $script:FLAG_ERROR=$true
                    return
                }
                }until($kp.KEY -eq "Enter")
        }
    }
    #if interactive query user for import scope
    #if only one key in file then no scope possible
    if($INTERACT -and $rx.dsp_key_array.Length -gt 1){
        $sis=Set-Import_scope
        if($sis -eq "q"){
            Write-Host "`nINFO:`tUser quit Import function`n"
            return
        }
        Write-Host ""
    }
    #filter dsp_key_array and data array
    $dsp_key_sel=@()
    $data_sel=@()
    if($PEQ -or $PRESET_KEY){
        if($PEQ -and $PQ2){
            $dsp_key_sel+=$rx.dsp_key_array|Where-Object{$_ -match "^Parametric Equalizer$"}
            $data_sel+=$rx.data|Where-Object{$_.path -match "\\Parametric Equalizer$"}
        }
        elseif($PEQ){
            $dsp_key_sel+=$rx.dsp_key_array|Where-Object{$_ -match "^Parametric"}
            $data_sel+=$rx.data|Where-Object{$_.path -match "\\Parametric"}
        }
        if($PRESET_KEY){
            foreach($pk in $PRESET_KEY){$dsp_key_sel+=$rx.dsp_key_array|Where-Object{$_ -match "^$pk$"}}
            foreach($key in $PRESET_KEY){$data_sel+=$rx.data|Where-Object{$_.Path -match "\\$key$"}}
        }
        #display selected keys
        $display_keys=($dsp_key_sel.ForEach{$_})-join "`n`t"
        Write-Host "INFO:`tSelected DSP keys to import:`n`t$display_keys"
    }
    #no filtering so just load all
    else{
        $dsp_key_sel=$rx.dsp_key_array
        $data_sel=$rx.data
        Write-Host "`nINFO:`tAll DSP keys are selected for import"
    }
    if($PQ2 -and $dsp_key_sel -match "Parametric Equalizer 2"){
        Write-Host "WARN:`tParameter PQ2 can only be used with import selections that do not include PEQ2 keys"
        Write-Host "INFO:`tSettings for Parametric Equalizer will NOT be imported to Parametric Equalizer 2"
    }
    if($PQ2 -and -not($dsp_key_sel -match "Parametric Equalizer 2")){
        Write-Host "WARN:`tParameter PQ2 set!"
        Write-Host "INFO:`tSettings for Parametric Equalizer will be imported to Parametric Equalizer 2"
        $data_sel|ForEach-Object{$_.Path=$_.Path -replace "Parametric Equalizer$","Parametric Equalizer 2"}
        $dsp_key_sel=$dsp_key_sel -replace "Parametric Equalizer$","Parametric Equalizer 2"
    }
    Write-Host ""
    #ask if user wants to change preset name if interactive
    if($INTERACT){
        Write-Host "QUERY:`tInput a new DSP preset name for import <or> Keep the DSP preset name: $file_preset"
        $script:PRESET=Read-Host -Prompt "INPUT:`tInput new [DSP preset name] <or> Press [ENTER] to keep existing name"
    }
    #check for user modification to import paths 
    if($PRESET -and $PRESET -ne $file_preset){
        Write-Host "`nINFO:`tDSP preset renamed: $PRESET"
        $preset_root=$preset_root -replace "Presets\\[^\\]+","Presets\$PRESET"
        $path_mod=$true
    }
    Write-Host ""
    #change paths if required
    if($path_mod){
        $data_sel|ForEach-Object{$_.Path=$_.Path -replace "^.+\sPresets\\.+\\","$preset_root\"}
        if($VERBOSE){Write-Host "INFO:`tThe import path for DSP key(s) has been modified to:`n`t$preset_root`n"}
    }
    #check for existing registry paths
    if(Test-Path $preset_root){$query_msg="MODIFY"}else{$query_msg="CREATE"}
    $p_name=$preset_root -replace "^.+\\"
    Write-Host "INFO:`tSearching registry for MC$mcv preset: $p_name"
    if($query_msg -eq "MODIFY"){
        Write-Host "INFO:`tPreset found, it will be MODIFIED!`n" 
    }
    else{
        Write-Host "INFO:`tPreset not found, it will be CREATED.`n"
    }
    foreach($key in $dsp_key_sel){
        if(Test-Path "$preset_root\$key"){
            $existing_key=$true
            Write-Host "WARN:`tProperties will be overwritten for key: $key"
        }
        else{
            Write-Host "INFO:`tProperties will be written to new key:  $key"
        }
    }
    Write-Host ""
    #if path already exists, warn or else create - give user go/nogo first
    if($existing_key){
        Write-Host "WARN:`tReady to MODIFY keys in MC$mcv preset: $p_name`nWARN:`tProperties will be overwritten in existing keys!"
        Write-Host "WARN:`tRecommend that a backup is created before continuing.`n"
    }
    else{
        Write-Host "INFO:`tReady to CREATE keys in MC$mcv preset: $p_name`n"
    }
    do{
        Write-Host "QUERY:`tPress [Q] to QUIT <or> Press [ENTER] to $query_msg $p_name`:" -NoNewline
        $kp=[console]::Readkey($true)
        Write-Host "`r"
        if($kp.KeyChar -eq "q"){
            Write-Host "INFO:`tUSER quit Import function"
            return
        }
    }until($kp.KEY -eq "Enter")
    Write-Host "`nINFO:`tWriting properties to MC$mcv DSP preset: $p_name`n"
    try{
        if(-not(test-path $preset_root)){
            if($VERBOSE){Write-Host "INFO:`tCreating key: $preset_root"}else{Write-Host "INFO:`tCreating key: $p_name"}
            New-Item -Path $preset_root |Out-Null
        }
        #Process array and write registry keys and properties
        Write-Host ""
        $data_sel.foreach{
            #only display path and number of properties once for each key
            if($prev_key -ne $_.Path){
                $pff=$_.Path
                $property_count=($data_sel|Where-Object{$_.path -eq $pff}).name.count
                if($VERBOSE){$wk_msg=$pff}else{$wk_msg=$pff -replace "^.+\\"}
                if(Test-Path $_.Path){
                    Write-Host "INFO:`tOverwriting key: $wk_msg"
                }
                else{
                    New-Item -Path $_.Path |Out-Null
                    Write-Host "INFO:`tCreating key: $wk_msg"
                }
                Write-Host "`t`tWriting $property_count properties to key"
            }
            if($VERBOSE){
                $property_name="`t`t" + $_.Name
                Write-Host $property_name
            }
            New-ItemProperty -Path $_.Path -Name $_.Name -Value $_.Value -PropertyType $_.Type -Force |Out-Null
            $prev_key=$_.Path
        }
    }#end of try
    catch{
        $PSerror=$error[0]
        Write-Host "ERROR:`tFailed to import MC$mcv DSP preset: $p_name`n`tError message is:`n`t$PSerror" -ForegroundColor Red
        $script:FLAG_ERROR=$true
        return
    }
    Write-Host "`nINFO:`tCompleted import of MC$mcv DSP preset: $p_name"
}#End of Function Import-DSP


Function Show-Preset{
    #Displays either list of presets or the keys within a preset
    #Parameter: $key selects preset key listing
    param([switch]$key)
    $mcv=Set-Preset_mcv -mc_msg "Please select a Media Center version to list from"
    if(-not$mcv){return}
    if($key){
        Write-Host "INFO:`tBeginning listing of Media Center $mcv DSP Preset keys"
        $valid_preset=Set-Valid_preset -mcv $mcv -vp_msg "to list keys for"
        if(-not$valid_preset){return}
        $reg_path="HKCU:\Software\J. River\Media Center $mcv\Zone Presets\$valid_preset"
        $info_msg="`nINFO:`t$valid_preset keys:"
        $errmsg="`nERROR:`tNo registry keys found for Media Center $mcv DSP Preset: $valid_preset"
    }
    else{
        $reg_path="HKCU:\Software\J. River\Media Center $mcv\Zone Presets"
        Write-Host "INFO:`tBeginning listing of Media Center $mcv DSP Presets"
        $info_msg="`nINFO:`tMC$mcv DSP Preset(s):"
        $errmsg="`nERROR:`tNo Media Center $mcv DSP presets were found!"
    }
    #Display either preset or keys
    If(Test-Path $reg_path){
        $key_list=(Get-ChildItem $reg_path -name) -join "`n`t"
        Write-Host $info_msg
        Write-Host "`n`t$key_list"
        Write-Host "`nINFO:`tCompleted listing."
    }
    else{
        Write-Host $errmsg -ForegroundColor Red
        $script:FLAG_ERROR=$true
    }
    return
}


Function Remove-Preset{
    #function deletes selected preset from registry
    $mcv=Set-Preset_mcv -mc_msg "Please select a Media Center version to delete preset from"
    if(-not$mcv){return}
    #set valid preset name for deletion
    Write-Host "INFO:`tBeginning Media Center $mcv DSP preset delete"
    $valid_preset=Set-Valid_preset -mcv $mcv -vp_msg "to delete"
    if(-not$valid_preset){return}
    Write-Host "`nQUERY:`tPlease confirm deletion of MC$mcv DSP preset: $valid_preset"
    Write-Host "`t[Y] Yes" -ForegroundColor Yellow -NoNewline
    Write-Host "`t[N] No (default is `"Y`"):" -NoNewline
    do{
        $kp=[console]::Readkey($true)
        if($kp.KeyChar -eq "n"){
            Write-Host "`n"
            Write-Host "INFO:`tCancelled DSP preset delete"
            return
        }
    }until($kp.KEY -eq "Enter" -or $kp.KeyChar -eq "y")
    Write-Host "`n"
    try{
        Remove-Item -Path "HKCU:\Software\J. River\Media Center $mcv\Zone Presets\$valid_preset" -Force -Recurse
        Write-Host "`nINFO:`tDeleted Media Center $mcv DSP preset $valid_preset"
    }
    catch{
        $PSerror=$error[0]
        Write-Host "ERROR:`tFailed to delete Media Center $mcv DSP preset $valid_preset" -ForegroundColor Red
        Write-Host "ERROR:`tError message is:`n`t$PSerror" -ForegroundColor Red
        $script:FLAG_ERROR=$true
        return
    }
}#end of function Remove-Preset


Function Invoke-Functions{
    #controls script execution and parses command line switches, ask for user input if none specified
    #Commandline switches: -h, -help, -export, -import, -peq, -list_presets, -list_keys, -list_file
    #parse help switch
    if($HELP){
        $help_file="$PSScriptRoot\MC-DSP-HELP.txt"
        If(Test-Path $help_file){
            Get-Content $help_file|more
        }
        else{
            Write-Host "ERROR:`tMC-DSP-HELP.txt is missing from MC-DSP folder!" -ForegroundColor Red
            Write-Host "WARN:`tCannot display Help file, please copy MC-DSP-HELP.txt to MC-DSP folder"
            Write-Host "HELP:`tUse Get-Help .\MC-DSP.ps1 to get Powershell Help for MC-DSP"
        }
        return
    }
    #parse info switches

    if($LIST_PRESETS){
        Show-Preset
        return
    }
    if($LIST_KEYS){
        Show-Preset -key
        return
    }
    if($LIST_FILE){
        Write-Host "INFO:`tBeginning listing of a MC-DSP xml file"
        Read-XML_file
        return
    }
    if($DELETE_PRESET){
        Remove-Preset
        return
    }
    #parse action switches if none then go interactive
    if($EXPORT){
        Write-Host "INFO:`tBeginning export of Media Center DSP preset(s) to file"
        Set-Export_params
        return
    }
    if($IMPORT){
        Write-Host "INFO:`tBeginning import of Media Center DSP presets from file"
        Import-DSP
        return
    }
    #display menu for function selection
    if(-not $EXPORT -and -not $IMPORT){
$H_msg=@"
`n
[H] HELP`tPress [H] to display MC-DSP Help file
[E] EXPORT`tPress [E] to export DSP preset to file
[I] IMPORT`tPress [I] to import DSP preset from a MC-DSP XML file
[P] PRESET list`tPress [P] to list saved DSP preset names
[D] DEL preset`tPress [D] to delete a preset
[K] KEY list`tPress [K] to list sub-keys of a DSP preset
[F] FILE info`tPress [F] to display summary of the content in a MC-DSP XML file
[Q] QUIT`tPress [Q] to exit MC-DSP
[?] Menu Help`tPress [?] to display help about menu choices
(default)`tPress [ENTER] to accept default choice of QUIT
"@
        
        do{
            Write-Host "`n`tMC-DSP MAIN MENU - Press a [key] to select a function`n"
            Write-Host "[H] HELP`t[E] EXPORT`t[I] IMPORT`t[P] PRESET list`t[D] DEL preset`n[K] KEY list`t[F] FILE info`t" -NoNewline
            Write-Host "[Q] QUIT`t" -ForegroundColor Yellow -NoNewline
            Write-Host "[?] Menu Help`t(default is `"Q`"):" -NoNewline
            $kp=[console]::Readkey($true)
            Write-Host "`r"
            if($kp.key -eq "Enter"){$script:INTERACT=$false;return}
        
            switch ($kp.KeyChar) {
                "h" {$vc=$true;$script:HELP=$true;break}
                "d" {$vc=$true;$script:DELETE_PRESET=$true;break}
                "e" {$vc=$true;$script:EXPORT=$true;break}
                "i" {$vc=$true;$script:IMPORT=$true;break}
                "p" {$vc=$true;$script:LIST_PRESETS=$true;break}
                "k" {$vc=$true;$script:LIST_KEYS=$true;break}
                "f" {$vc=$true;$script:LIST_FILE=$true;break}
                "q" {$vc=$true;$script:INTERACT=$false;return}
                "?" {Write-Host $H_msg;break}
                Default {Write-Host "WARN:`tInvalid choice!"}
            }
        }until($vc)
        $script:INTERACT=$true
        #Interactive mode ignore (clear) any command line modifier parameters apart from MC_VER and XML_PATH
        Clear-Variable -Name FILE,PEQ,PQ2,PRESET,PRESET_KEY -Scope Script
        #now we have some parameters start Invoke-Functions again
        Write-Host "`n"
        Invoke-Functions
    }
}#End of Function Invoke-Functions


#start of script MC-DSP.ps1
#use try-finally to handle ctrl-C exit or unhandled error
try{
    #add forms for save file and browse file
    Add-Type -AssemblyName System.Windows.Forms
    Write-Host "`n"
    Invoke-Functions
    if($INTERACT){
        do{
            #clear all script level variables and start again
            Clear-Variable -Scope Script -Name DELETE_PRESET,EXPORT,FILE,HELP,IMPORT,FLAG_ERROR,LIST_FILE,LIST_KEYS,LIST_PRESETS,
            PEQ,PQ2,PRESET,PRESET_KEY -ErrorAction SilentlyContinue
            Invoke-Functions
        }until(-not$INTERACT)
    }
    Write-Host "`n"
    $eos=$true
}
finally{
    if($FLAG_ERROR){
        Write-Host "WARN:`tMC-DSP was terminated by an error! Please review error message`n"
    }
    if(-not$eos){
        Write-Host "`nWARN:`tMC-DSP was terminated by ctrl-C or a fatal unhandled error`n"
}
}
#End of script MC-DSP.ps1